# (c) Copyright 2011 Cloudera, Inc. All rights reserved.

import os
import logging
import threading
import time
import avro.ipc
import avro.protocol

from cmf.monitor.base import DebugStatusReport
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)

_CONN_FAILED_LOG_DELAY = 10 * 60
THROTTLED_LOG = ThrottlingLogger(LOG, _CONN_FAILED_LOG_DELAY)

FIREHOSE_MESSAGE_PROTOCOL = avro.protocol.parse(
  file(os.path.join(os.path.dirname(__file__), "cmon_agent.avpr")).read())

class Firehose(object):
  """This class is used for communication with a firehose. It is thread safe."""

  def __init__(self, name, roletype, address, port):
    self.name = name
    self.roletype = roletype
    self._address = address
    self._port = port
    self._requestor = None
    self._transceiver = None
    self._shutdown = False
    self._lock = threading.Lock()

  def set_connection_info(self, address, port):
   try:
      self._lock.acquire()
      self._set_connection_info(address, port)
   finally:
     self._lock.release()

  def _set_connection_info(self, address, port):    
    if self._address == address and self._port == port:
      return
    if self._shutdown:
      LOG.info("Ignoring connection update on shutdown firehose.")
      return
    if address is not None and port is not None:
      LOG.info("Updating firehose connection to %s:%s:" % (address, port))
    else:
      LOG.info("Updating firehose connection to be unconfigured.")
    self._address = address
    self._port = port
    self._reset()

  def send(self, messages):
   try:
      self._lock.acquire()
      self._send(messages)
   finally:
     self._lock.release()

  def _send(self, messages):
    """Sends messages to the firehose. Returns True on success."""
    if self._shutdown:
      LOG.info("Ignoring send request on shutdown firehose.")
      return False
    try:
      if self._requestor is None:
        self._transceiver = avro.ipc.HTTPTransceiver(self._address,
                                                     self._port)
        self._requestor = avro.ipc.Requestor(FIREHOSE_MESSAGE_PROTOCOL,
                                             self._transceiver)

      self._requestor.request('sendAgentMessages', dict(messages=messages))
      return True
    except Exception:
      THROTTLED_LOG.exception("Error sending messages to firehose: " + self.name)
      self._reset()
      return False

  def _reset(self):
    if self._transceiver is not None:
      try:
        self._transceiver.close()
      except Exception:
        THROTTLED_LOG.exception("Error closing transciever: " + self.name)
    self._transceiver = None
    self._requestor = None

  def shutdown(self):
   try:
      self._lock.acquire()
      self._shutdown = True
      self._reset()
   finally:
     self._lock.release()

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=[('address', self._address),
                                           ('port', self._port),
                                           ('shutdown', self._shutdown)])

  def __str__(self):
    return "Firehose(%s, %s, %s, %s)" % (self.name,
                                         self.roletype,
                                         self._address,
                                         self._port)
