# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

"""
The host monitor
"""

import simplejson as json
import os
import psutil
import logging

from cmf.monitor import constants
from cmf.monitor import schema
from cmf.monitor.abstract_monitor import AbstractMonitor
from cmf.monitor.host.filesystem_map import FilesystemMap
from cmf.monitor.host_update import HostUpdate
from cmf.monitor.proc_metrics_utils import get_swap_out
from cmf.monitor.wakeable_thread import WakeableThread
from cmf.monitor.host.network_interfaces import NetworkInterfaces
from cmf.monitor.host.disk_devices import DiskDevices
from cmf.monitor.host.dns_names import DnsNames
from cmf.monitor.host.tcp_metrics import TcpMetrics
from cmf.monitor.host.filesystems import Filesystems
from cmf.monitor.proc_metrics_utils import get_fd_count, get_fd_max
from cmf.throttling_logger import ThrottlingLogger

# TODO(bc): We should get collection interval from SCM server
_COLLECTION_INTERVAL_SEC = 60

NEVER = "-2.0"

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class HostMonitor(AbstractMonitor):

  def __init__(self, daemon, registry):
    AbstractMonitor.__init__(self, daemon, registry, None, None)
    self._filesystems = Filesystems(daemon.get_filesystem_map())
    self._disk_devices = DiskDevices()
    self._tcp_metrics = TcpMetrics()
    self._supervisor_successes = 0
    self._supervisor_latency_sec = 0.0
    try:
      self._network_interfaces = NetworkInterfaces()
    except IOError, ioex:
      LOG.warning("Unable to create network interface stats reader, "
                  "agent will not report NIC metrics. errno=%s" % (ioex.errno,))
      self._network_interfaces = None
    self._last_clock_offset = None
    self._host_dir_monitors = [ ]

    if self._daemon.get_agent() is not None:
      self.create_dir_monitor(self._daemon.get_agent().logdir,
                              self._host_dir_monitors)
      self.create_dir_monitor(self._daemon.get_agent().dir,
                              self._host_dir_monitors)
      self.create_dir_monitor(self._daemon.get_agent().parcel_dir,
                              self._host_dir_monitors)
    self.create_dns_monitor()
    self._monitor_dns = False

  def __str__(self):
    return "HostMonitor"

  def make_collection_thread(self):
    return WakeableThread(
          "Monitor-%s" % (self,),
          self.collect)

  def collect(self):
    """The loop content in the collection thread."""
    self._filesystems.collect()
    self._disk_devices.collect()
    if self._network_interfaces is not None:
      self._network_interfaces.collect()
    for mon in self._host_dir_monitors:
      mon.collect()
    if self._monitor_dns and self._dns_names is not None: 
      self._dns_names.collect()

  def get_host_update(self):
    """Returns a host update to be sent to the firehose"""
    agent_log_dir = None
    agent_process_dir = None
    agent_parcel_dir = None
    if self._daemon.get_agent() is not None:
      agent_log_dir = self._daemon.get_agent().logdir
      agent_process_dir = self._daemon.get_agent().dir
      agent_parcel_dir = self._daemon.get_agent().parcel_dir
    update = HostUpdate(agent_log_dir, agent_process_dir, agent_parcel_dir)

    if not self._last_clock_offset is None:
      update.add_metric(schema.CLOCK_OFFSET,
                        self._last_clock_offset)

    for fs_update in self._filesystems.get_fs_updates():
      update.add_fs_update(fs_update)

    for disk_update in self._disk_devices.get_disk_updates():
      update.add_disk_update(disk_update)

    if self._network_interfaces is not None:
      for nic_update in self._network_interfaces.get_nic_updates():
        update.add_nic_update(nic_update)

    for mon in self._host_dir_monitors:
      update.add_dir_update(mon.get_host_dir_update())

    # Add host update metrics
    # interval=None means "averaged since last call"
    # Could use cpu_times instead to get different contexts
    # (user, system, etc.)
    update.add_metric(schema.CPU_PERCENT, psutil.cpu_percent(interval=None))

    load_avg = os.getloadavg()
    update.add_metric(schema.LOAD_1, load_avg[0])
    update.add_metric(schema.LOAD_5, load_avg[1])
    update.add_metric(schema.LOAD_15, load_avg[2])

    cpu_time = psutil.cpu_times()
    update.add_metric(schema.TOTAL_CPU_USER, cpu_time.user)
    update.add_metric(schema.TOTAL_CPU_SYSTEM, cpu_time.system)
    update.add_metric(schema.TOTAL_CPU_IDLE, cpu_time.idle)
    update.add_metric(schema.TOTAL_CPU_NICE, cpu_time.nice)
    update.add_metric(schema.TOTAL_CPU_IOWAIT, cpu_time.iowait)
    update.add_metric(schema.TOTAL_CPU_IRQ, cpu_time.irq)
    update.add_metric(schema.TOTAL_CPU_SOFT_IRQ, cpu_time.softirq)

    agent_process = psutil.Process(os.getpid())
    agent_cpu_time = agent_process.get_cpu_times()
    if agent_cpu_time is not None:
      update.add_metric(schema.AGENT_CPU_USER, agent_cpu_time.user)
      update.add_metric(schema.AGENT_CPU_SYSTEM, agent_cpu_time.system)
    rss, vms = agent_process.get_memory_info()
    update.add_metric(schema.AGENT_PHYSICAL_MEMORY_USED, rss)
    update.add_metric(schema.AGENT_VIRTUAL_MEMORY_USED, vms)
    update.add_metric(schema.AGENT_FD_OPEN, get_fd_count(os.getpid()))
    update.add_metric(schema.AGENT_FD_MAX, get_fd_max(os.getpid()))

    try:
      supervisord_process = self._daemon.get_agent().supervisor_psutil_process
      supervisord_pid = self._daemon.get_agent().supervisor_pid
      supervisord_cpu_time = supervisord_process.get_cpu_times()
      if supervisord_cpu_time is not None:
        update.add_metric(schema.SUPERVISORD_CPU_USER, supervisord_cpu_time.user)
        update.add_metric(schema.SUPERVISORD_CPU_SYSTEM, supervisord_cpu_time.system)
      rss, vms = supervisord_process.get_memory_info()
      update.add_metric(schema.SUPERVISORD_PHYSICAL_MEMORY_USED, rss)
      update.add_metric(schema.SUPERVISORD_VIRTUAL_MEMORY_USED, vms)
      update.add_metric(schema.SUPERVISORD_FD_OPEN, get_fd_count(supervisord_pid))
      update.add_metric(schema.SUPERVISORD_FD_MAX, get_fd_max(supervisord_pid))
    except Exception, e:
      LOG.info("Couldn't get supervisord metrics: %s" % (e,))

    if self._daemon.get_agent() is not None:
      supervisor_statistics = self._daemon.get_agent().supervisor_statistics
      supervisor_successes = supervisor_statistics["successes"]
      supervisor_successes_delta = supervisor_successes - self._supervisor_successes
      self._supervisor_successes = supervisor_successes
      supervisor_latency_sec = supervisor_statistics["latency_sec"]
      supervisor_latency_delta = supervisor_latency_sec - self._supervisor_latency_sec
      self._supervisor_latency_sec = supervisor_latency_sec
      if supervisor_successes_delta > 0:
        update.add_metric(schema.SUPERVISORD_LATENCY, supervisor_latency_delta / supervisor_successes_delta);
      update.add_metric(schema.SUPERVISORD_FAILURES, supervisor_statistics["failures"]);

    cached = psutil.cached_phymem()
    buffers = psutil.phymem_buffers()
    really_used = psutil.used_phymem() - cached - buffers
    virtmem_usage = psutil.virtmem_usage()
    update.add_metric(schema.PHYSICAL_MEMORY_USED, really_used)
    update.add_metric(schema.PHYSICAL_MEMORY_CACHED, cached)
    update.add_metric(schema.PHYSICAL_MEMORY_BUFFERS, buffers)
    update.add_metric(schema.PHYSICAL_MEMORY_TOTAL, psutil.TOTAL_PHYMEM)
    update.add_metric(schema.SWAP_TOTAL, virtmem_usage.total)
    update.add_metric(schema.SWAP_FREE, virtmem_usage.free)
    update.add_metric(schema.SWAP_USED, virtmem_usage.used)
    swap_out = get_swap_out()
    if swap_out is not None:
      update.add_metric(schema.SWAP_OUT, swap_out)

    self._tcp_metrics.add_tcp_metrics(update)

    if self._monitor_dns and self._dns_names is not None:
      self._dns_names.add_dns_metrics(update)

    return update

  def start_collecting(self, latch):
    if self._collection_thread is not None:
      self._collection_thread.wake(latch)

  def record_heartbeat_timestamps(self,
                                  ts_agent_send,
                                  ts_agent_recv,
                                  ts_master_send,
                                  ts_master_recv):
    if ts_agent_send >= 0 and \
       ts_agent_recv >= 0 and \
       ts_master_send >= 0 and \
       ts_master_recv >= 0:
      processing_time = ts_master_send - ts_master_recv
      total_time = ts_agent_recv - ts_agent_send
      latency = (total_time - processing_time) / 2

      expected_agent_recv = ts_master_send + latency
      # Take absolute value because charts don't currently handle negative values
      # and offset can be positive or negative (OPSAPS-7789)
      self._last_clock_offset = \
          abs(expected_agent_recv - ts_agent_recv)

  def update_host_config(self, host_config):
    if (host_config.has_key('host_network_interface_collection_filter')):
      if self._network_interfaces is not None:
        self._network_interfaces.set_iface_exclude_regex(
             host_config['host_network_interface_collection_filter'])
    if (host_config.has_key('host_disk_collection_filter')):
      self._disk_devices.set_device_exclude_regex(
           host_config['host_disk_collection_filter'])
    if (host_config.has_key('host_fs_collection_filter')):
      self._filesystems.set_fs_exclude_regex(
           host_config['host_fs_collection_filter'])
    self._monitor_dns = self.update_monitor_dns(host_config)

  def is_ready_to_report(self):
    return True

  def get_subject_type(self):
    return "HOST"

  def create_dns_monitor(self):
    self._dns_names = None
    try:
      mgmt_home = self._daemon.get_agent().supervisee_environment["MGMT_HOME"]
      agent_jar_name = 'agent-' + constants.AGENT_VERSION + '.jar'
      agent_jar_path = mgmt_home + '/lib/' + agent_jar_name
      if os.path.exists(agent_jar_path):
        self._dns_names = DnsNames(agent_jar_path)
      else:
        THROTTLED_LOG.error("Jar %s not found." % (agent_jar_name))
    except:
      THROTTLED_LOG.error("Unable to get MGMT_HOME from daemon.agent.supervisee_environment")

  def update_monitor_dns(self, host_config):
    dns_resolution_enabled = \
        host_config.has_key('host_dns_resolution_enabled') and \
        host_config['host_dns_resolution_enabled'] == 'true'
    dns_resolution_duration_enabled = False
    if (host_config.has_key('host_dns_resolution_duration_thresholds')):
      try:
        duration_thresholds = json.loads(host_config['host_dns_resolution_duration_thresholds'])
        dns_resolution_duration_enabled = \
            (duration_thresholds['warning'] != NEVER) or \
            (duration_thresholds['critical'] != NEVER)
      except:
        THROTTLED_LOG.log("Unable to parse host_dns_resolution_duration_thresholds : %s" \
                          % (host_config['host_dns_resolution_duration_thresholds']))
    return dns_resolution_enabled or dns_resolution_duration_enabled
